// Generated by Together
package edu.cmu.paid.learning.dataminer.developer.NaiveBayes;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.util.StringTokenizer;
import java.io.*;
import java.rmi.*;
import edu.cmu.paid.learning.dataminer.util.MLCDriver;
import edu.cmu.paid.learning.dataminer.developer.RemoteModelBuilder;
import edu.cmu.paid.learning.dataminer.developer.BehaviorModel;


/**
 * NBBuilder is an implementation of RemoteModelBuilder that runs as a remote
 * server that can be contacted by RemoteDeveloper for model development.
 *
 * @author Rudy Setiawan, PAID Learning Team, Carnegie Mellon University.
 * @version %I%, %G%
 */
public class NBBuilder extends java.rmi.server.UnicastRemoteObject implements RemoteModelBuilder 
{
  private String hostName;
  private String path;
  private String name = "NaiveBayes";


  /**
   * main function to start the ID3Builder as a server. The RMI registry must
   * be running before ID3Builder is started. Two arguments are needed to start
   * the ID3Builder: the RMI registry to which the ID3Builder will bind to and
   * the path where the inducer is located.
   */
  public static void main(String[] args)
  {
    String path, hostName;

    if(args.length !=2)
      return;
    else
      {
	hostName = args[0];
        path = args[1];
      }
    
    try
      {
	NBBuilder builder = new NBBuilder(hostName, path);
	builder.connect();
      }
    catch(Exception e)
      {
	System.out.println("Exception " + e + " caught.");
	e.printStackTrace();
      }
  }


  /**
   * Constructor for NBBuilder takes a couple of arguments as given below.
   *
   * @param hostName the RMI registry to which the NBBuilder will bind to
   * @param path the directory path where the Naive Bayes inducer is located
   */
  public NBBuilder(String hostName, String path) throws java.rmi.RemoteException
    {
      this.hostName = hostName;
      this.path = path;

      if(!path.endsWith(File.separator))
	path = path + File.separator;
    }


  /** 
   * Connects to the RMI registry.
   */
  private void connect()
    { 
      try
        {
          Naming.rebind("rmi://" + hostName + "/" + name, this);

          System.out.println("NBBuilder " + name + " is ready");
        }
      catch(Exception e)
        {
          System.out.println("Exception " + e + " caught when connecting RemoteBuilder");
        }
    }


  /**
   * Builds a BehaviorModel based on the request for model development.
   *
   * @param request the request for model development.
   * @return the BehaviorModel produced.
   * @see NBModel
   */
  public BehaviorModel build(Hashtable request) throws java.rmi.RemoteException
    {
      System.out.println("DEBUG: NaiveBayes Receiving request to build a model");
      Hashtable metadata = (Hashtable) request.get("metadata");
      Vector trainingSet = (Vector) request.get("training set");

      String timestamp = (new Long(System.currentTimeMillis())).toString();
      String datasetName = path + name;

      MLCDriver mlc = new MLCDriver(path, name);
      mlc.setMLCFile(metadata, trainingSet);
      String categorizerFilename = datasetName + ".nbmodel";
      System.out.println("DEBUG: categorizerFilename is " + categorizerFilename);
      String command = path + "NBInducer " + datasetName + " " + categorizerFilename;
      System.out.println("DEBUG: Command to be executed is " + command);
      mlc.runInducer(command);
      Hashtable categorizer = getCategorizer(metadata, categorizerFilename);

      cleanUp(datasetName);

      BehaviorModel model = new NBModel(categorizer);

      System.out.println("DEBUG: Finished generating NBModel. About to deliver");
      return model;      
    }


  /**
   * Cleans up the mess after generating the BehaviorModel.
   * 
   * @param dName the dataset's path and name
   */
  private void cleanUp(String dName)
    {
      String metaFilename = dName + ".names";
      String trainingFilename = dName + ".data";
      String categorizerFilename = dName + ".id3model";

      (new File(metaFilename)).delete();
      (new File(trainingFilename)).delete();
      (new File(categorizerFilename)).delete();
    }


  /**
   * Gets the categorizer produced by the inducer to be used in initializing
   * an instance of NBModel.
   *
   * @param metaData the metadata of the learning task
   * @param categorizerFile the name of categorizer file to be generated by 
   *                        the inducer.
   * @return the categorizer
   */
  private Hashtable getCategorizer(Hashtable metaData, String inFile)
    {
      Hashtable toBeReturned = new Hashtable();;
      Vector firstBuffer = new Vector();
    
      // parsing MLC++ NaiveBayes output
      // first, just getting rid of "junk" lines
      try
        {
	  FileInputStream fis = new FileInputStream(inFile);
	  BufferedReader br = new BufferedReader(new InputStreamReader(fis));
	
	  while(br.ready())
	    {
	      String aLine = br.readLine();
	      aLine = aLine.trim();
	    
	      if((aLine.startsWith("Value")) || (aLine.startsWith("Attribute")) || (aLine.startsWith("Label")) || (aLine.startsWith("Categorizing")))
	        {
		  firstBuffer.addElement(aLine);
	        }
	    }
        }
      catch(FileNotFoundException fnfe)
        {
	  System.out.println("FileNotFoundException " + fnfe);
        }
      catch(IOException ioe)
        {
	  System.out.println("IOException " + ioe);
        }
    
      // second level parsing: dividing result of level 1 into three parts:
      //   the value counter (for discrete vars), attribute count, and normal
      //   distribution parameter (for continous vars)
    
      int part = 0;
      String partID = null;
      Vector valCounter = new Vector();
      Vector attCount = new Vector();
      Vector ndParam = new Vector();
    
      Vector varNameList = (Vector) metaData.get("Independent");    
      Vector depVarValues = (Vector) metaData.get("Dependent");
    
      for(int i=0; i<firstBuffer.size(); i++)
        {
	  String item = (String) firstBuffer.elementAt(i);
	
	  if(item.equals("Value counters:"))
	    part = 1;
 	  else if(item.equals("Attribute counts"))
	    part = 2;
	  else if(item.startsWith("Categorizing uses Normal"))
	    part = 3;
	  else
	    {
	      if(part == 1)
	        {
		  int vcAttribIndex = 0;
		  StringTokenizer valCountHeadTok = new StringTokenizer(item, " :");
		  valCountHeadTok.nextToken();
		  vcAttribIndex = Integer.valueOf(valCountHeadTok.nextToken()).intValue();
		
		  Hashtable vcValCounts = new Hashtable();
		  int vcLabelIndex = 0;
		  boolean vcNotDone = true;
		
		  while (vcNotDone)
		    {
		      String nextItem = (String) firstBuffer.elementAt(i+1);
		    
		      // if continuous, insert null to valCounter
		      if(nextItem.startsWith("Attribute"))
		        {
			  // System.out.println("inserted: " + vcValCounts.toString());
			  valCounter.insertElementAt(vcValCounts, vcAttribIndex);
			  vcNotDone = false;
		        }
		      else
		        {
			  i++;
			  nextItem = (String) firstBuffer.elementAt(i);
			  // if starts with "Label", get the index
			  if (nextItem.startsWith("Label"))
			    {
			      StringTokenizer vcLabelTok = new StringTokenizer(nextItem, " ,");
			      vcLabelTok.nextToken();
			      vcLabelIndex = Integer.valueOf(vcLabelTok.nextToken()).intValue();
			    }
			  // else if starts w/ Value counts, get frequency vector
			  else if (nextItem.startsWith("Value counts"))
			    {
			      StringTokenizer vcValueCountsTok1 = new StringTokenizer(nextItem, ":");
			      vcValueCountsTok1.nextToken();
			      String vcValueString = (vcValueCountsTok1.nextToken()).trim();
			      StringTokenizer vcValueCountsTok2 = new StringTokenizer(vcValueString, " ,");
			    
			      Hashtable frequencyValues = new Hashtable();
			    
			      // Vector frequencyValues = new Vector();
			    
			      Integer vcValueCountsItem = Integer.valueOf(vcValueCountsTok2.nextToken());
			    
			      frequencyValues.put("?", vcValueCountsItem);
			      int counter = 0;
			    
			      String indVarName = (String) varNameList.elementAt(vcAttribIndex);
			    
			      Vector depVarVals = (Vector) metaData.get(indVarName);
			    
			      while(vcValueCountsTok2.hasMoreTokens())
			        {
				  vcValueCountsItem = Integer.valueOf(vcValueCountsTok2.nextToken());
				
				  String valName = (String) depVarVals.elementAt(counter);
				  counter++;
				
				  frequencyValues.put(valName, vcValueCountsItem);
			        }
			    
			      String label;

			      if(vcLabelIndex == -1)
			        label = "?";
			      else
			        label = (String) depVarValues.elementAt(vcLabelIndex);
			    
			      vcValCounts.put(label, frequencyValues);
			    }
		        }
		    }
	        }
	      else if(part == 2)
	        attCount.addElement(item);
	      else if(part == 3)
	        {
		  int ndpAttribIndex = 0;
		  StringTokenizer ndpHeadTok = new StringTokenizer(item, " :");
		  ndpHeadTok.nextToken();
		  ndpAttribIndex = Integer.valueOf(ndpHeadTok.nextToken()).intValue();
		
		  Hashtable ndpVars = new Hashtable();
		  int ndpLabelIndex = 0;
		  boolean ndpNotDone = true;
		
		  while (ndpNotDone)
		    {
		      String nextItem = null;
		    
		      if(i == firstBuffer.size()-1)
		        {
			  ndParam.insertElementAt(ndpVars, ndpAttribIndex);
			  break;
		        }
		    
		      nextItem = (String) firstBuffer.elementAt(i+1);
		    
		      if(nextItem.startsWith("Attribute"))
		        {
			  ndParam.insertElementAt(ndpVars, ndpAttribIndex);
			  ndpNotDone = false;
		        }
		      else
		        {
			  i++;
			  nextItem = (String) firstBuffer.elementAt(i);
			  StringTokenizer ndpLabelTok = new StringTokenizer(nextItem, " \t");
			  ndpLabelTok.nextToken();
			  ndpLabelIndex = Integer.valueOf(ndpLabelTok.nextToken()).intValue();
			
			  Hashtable vars = new Hashtable();
			
			  vars.put("mean", Double.valueOf(ndpLabelTok.nextToken()));
			  vars.put("variance", Double.valueOf(ndpLabelTok.nextToken()));
			
			  String label;
			
			  if(ndpLabelIndex == -1)
			    label = "?";
			  else
			    label = (String) depVarValues.elementAt(ndpLabelIndex);
			    
			  ndpVars.put(label, vars);
		        }
		    }
	        }
	    }
        }
    
      // calculate total count per dependent variable value
      Hashtable totalCount = new Hashtable();
    
      for(int i=0; i<valCounter.size(); i++)
        {
	  String attributeName = (String) varNameList.elementAt(i);
	  Hashtable item = (Hashtable) valCounter.elementAt(i);
	
	  if(!totalCount.isEmpty()) break;
	
	  if(!item.isEmpty())
	    {
	      Hashtable frequencyValues = (Hashtable) item.get("?");

	      // calculate total count
	      Vector indVarVals = (Vector) metaData.get(attributeName);

	      int total = 0;
	    
	      for(int k=0; k<indVarVals.size(); k++)
	        {
		  String indVarName = (String) indVarVals.elementAt(k);
		
		  total += ((Integer) frequencyValues.get(indVarName)).intValue();
	        }
	    
	      // System.out.println("For ?, total is " + total);
	    
	      totalCount.put("?", new Integer(total));
	    
	      for(int j=0; j<depVarValues.size(); j++)
	        {
		  String depVarName = (String) depVarValues.elementAt(j);
		
		  frequencyValues = (Hashtable) item.get(depVarName);  
		
		  // calculate total count
		  indVarVals = (Vector) metaData.get(attributeName);
		
		  total = 0;
		
		  for(int k=0; k<indVarVals.size(); k++)
		    {
		      String indVarName = (String) indVarVals.elementAt(k);
		    
		      total += ((Integer) frequencyValues.get(indVarName)).intValue();
		    }
		  totalCount.put(depVarName, new Integer(total));
	        }
	    }
        }
    
      // System.out.println("totalCount is " + totalCount.toString());
    
      for(int i=0; i<valCounter.size(); i++)
        {
	  String attributeName = (String) varNameList.elementAt(i);
	  Hashtable item = (Hashtable) valCounter.elementAt(i);
	
	  if (!item.isEmpty())
	    {
	      // deal with discrete attributes
	      Hashtable frequencyValues = (Hashtable) item.get("?");  
	    
	      Vector indVarVals = (Vector) metaData.get(attributeName);
	    
	      calculateProbability("?", totalCount, frequencyValues, indVarVals, "?");
	    
	      for(int k=0; k<indVarVals.size(); k++)
	        {
		  String indVarName = (String) indVarVals.elementAt(k);
		
		  calculateProbability("?", totalCount, frequencyValues, indVarVals, indVarName);
	        }
	    
	      for(int j=0; j<depVarValues.size(); j++)
	        {
		  String depVarName = (String) depVarValues.elementAt(j);
		
		  frequencyValues = (Hashtable) item.get(depVarName);  
		
		  indVarVals = (Vector) metaData.get(attributeName);
		
		  calculateProbability(depVarName, totalCount, frequencyValues, indVarVals, "?");
		
		  for(int k=0; k<indVarVals.size(); k++)
		    {
		      String indVarName = (String) indVarVals.elementAt(k);
		    
		      calculateProbability(depVarName, totalCount, frequencyValues, indVarVals, indVarName);
		    }
	        }
	    
	      // System.out.println(attributeName + ": " + item);
	    }
	  else
	    {
	      // deal with continuous attributes
	      item = (Hashtable) ndParam.elementAt(i);
	    
	      for(int j=0; j<depVarValues.size(); j++)
	        {
		  String depVarName = (String) depVarValues.elementAt(j);
		
		  Hashtable params = (Hashtable) item.get(depVarName);
		
		  double mean =  ((Double) params.get("mean")).doubleValue();
		  double std = Math.sqrt(((Double) params.get("variance")).doubleValue());
		
		  NormalDistribution nd = new NormalDistribution(mean, std);
		
		  item.put(depVarName, nd);
	        }
	    
	      // System.out.println(attributeName + ": " + item);
	    }
        }

      toBeReturned.put("metadata", metaData);
      toBeReturned.put("discrete", valCounter);
      toBeReturned.put("continuous", ndParam);
    
      return toBeReturned;
    }

  private void calculateProbability(String depVarName, Hashtable totalCount, Hashtable frequencyValues, Vector indVarVals, String varName)
    {
      double total = ((Integer) totalCount.get(depVarName)).doubleValue();
    
      if(total !=0)
        {
	  double frequency =  ((Integer) frequencyValues.get(varName)).doubleValue();
	
	  double probability = frequency/total;
	
	  frequencyValues.put(varName, new Double(probability));
        }
      else
        {
	  double probability = 1/((double) indVarVals.size()+ 1);
	
	  frequencyValues.put(varName, new Double(probability));
        }
    }
}




